from functools import reduce, wraps
import collections

def requires_context(fn):
    fn.requires_context = True
    return fn

def is_seq(v):
    return isinstance(v, collections.Iterable)

def is_pred(pred):
    def w1(fn):
        @wraps(fn)
        def w2(v):
            if not pred(v):
                raise ValueError("given value %r fails predicate %r" % (v, pred.__name__))
            return fn(v)
        return w2
    return w1

is_str = is_pred(lambda v: isinstance(v, str))

def do_all_if_tuple(fn):
    @wraps(fn)
    def wrapper(v):
        if isinstance(v, tuple):
            return tuple([fn(x) for x in v])
        return fn(v)
    return wrapper

@do_all_if_tuple
@is_str
def uppercase(v):
    return v.upper()

@do_all_if_tuple
@is_str
def lowercase(v):
    return v.lower()

@do_all_if_tuple
@is_str
def titlecase(v):
    return v.title()

class PipelineError(Exception):
    pass

def pipeline(*pline):
    """a wrapper around the typical list that helps with debugging
    
    usage: 
      description = {'foo': pipeline(123, et3.utils.lowercase)}
      et3.render(description)"""
    def wrapper(processor, item):
        try:
            return processor(item, pline)
        except Exception as err:
            def forn(x):
                if hasattr(x, '__name__'):
                    return 'fn:' + x.__name__
                return str(x)
            msg = "pipeline %r failed with: %s" % ([forn(v) for v in pline], err)
            # LOG.error(doi(item) + " - caught exception attempting to render: " + msg)
            raise PipelineError(msg)
    return wrapper



from .utils import is_seq
from collections import OrderedDict

EXCLUDE_ME = 0xDEADC0DE

def default_context(ctx=None):
    return {} if not isinstance(ctx, dict) else ctx

def do(item, segment, ctx=None):
    if callable(segment):
        if hasattr(segment, 'requires_context'):
            return segment(default_context(ctx), item)
        return segment(item)
    return segment

def doall(item, pipeline, ctx=None):
    for seg in pipeline:
        if isinstance(seg, tuple):
            item = tuple([do(item, subseg, ctx) for subseg in seg])
        else:
            item = do(item, seg, ctx)
    return item

def render_item(description, item, ctx=None):
    result = OrderedDict() if isinstance(description, OrderedDict) else {}
    for key, pipeline in description.items():
        if isinstance(pipeline, dict):
            sub_description = pipeline
            rendered = render_item(sub_description, item, ctx)
        elif isinstance(pipeline, list):
            rendered = doall(item, pipeline, ctx)
        elif callable(pipeline):
            rendered = pipeline(doall, item) # no ctx? I might yank callable pipelines ...
        else:
            raise AssertionError("render pipeline for item is an unhandled type: %r" % type(pipeline))
        if rendered == EXCLUDE_ME:
            # pipeline has indicated this key should be discarded from the results
            continue
        result[key] = rendered
    return result

def render(description, data, ctx=None):
    assert is_seq(data), "data must be a sequence of values, not %r" % type(data)
    return map(lambda item: render_item(description, item, ctx), data)



#!/usr/bin/env python
from setuptools import setup

VERSION = '1.5.1'
BASE_URL = 'https://bitbucket.org/lskibinski/et3'

setup(**{
    'name': 'et3',
    'version': VERSION,
    'description': 'Simple library for Extracting and Transforming data, third incarnation.',
    'url': BASE_URL,
    'download_url': BASE_URL + '/get/' + VERSION + '.tar.gz',
    'license': 'GPLv3',
    'author': 'Luke',
    'author_email': 'lsh-0@users.noreply.github.com',
    'test_suite': 'et3.tests',
    'packages': ['et3'],
    'install_requires': [], # see install.sh and requirements-dev.txt for local development
    'platforms': ['any'],
    'classifiers': [
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    ]
})



from functools import wraps

def lookup(data, path, default=0xDEADBEEF):
    if not (isinstance(data, dict) or isinstance(data, list)):
        raise ValueError("lookup context must be a dictionary or a list, got %r: %r" % (type(data), data))
    if not isinstance(path, str):
        raise ValueError("path must be a string, given %r", path)
    try:
        bits = path.split('.', 1)
        if len(bits) > 1:
            bit, rest = bits
        else:
            bit, rest = bits[0], []

        if isinstance(data, dict):
            val = data[bit]
        else:
            val = data[int(bit)]
        if rest:
            return lookup(val, rest, default)
        return val
    except KeyError:
        if default == 0xDEADBEEF:
            raise
        return default

def path(p, default=0xDEADBEEF):
    def fn(data):
        if p == None:
            return data
        return lookup(data, p, default)
    fn.__name__ = 'path'
    return fn

def val(v):
    "simple placeholder for the value coming from the list of items"
    return v
